
# include "..\include\child.h"

// HTTP error codes
#define HTTP_200	200
#define HTTP_301	301
#define HTTP_304	304
#define HTTP_400	400
#define HTTP_404	404
#define HTTP_405	405
#define HTTP_500	500
#define HTTP_501	501
#define HTTP_503	503

//
// define some well known mime types
//

struct {
  char *ext;
  char *filetype;
} extensions [] = {
  {"htm",  "text/html"    },
  {"html", "text/html"    },
  {"ico",  "image/x-icon" },
  {"gif",  "image/gif"    },
  {"jpg",  "image/jpeg"   },
  {"jpeg", "image/jpeg"   },
  {"png",  "image/png"    },
  {"zip",  "image/zip"    },
  {"gz",   "image/gz"     },
  {"tar",  "image/tar"    },
  {"txt",  "text/plain"   },
  {0,0}
};

//
// http 1.0 (and some 1.1) status codes
//

struct {
  int code;
  char *message;
} http_codes [] = {
  {HTTP_200,  "OK"                  },
  {HTTP_301,  "Moved Permanently"   },
  {HTTP_304,  "Not Modified"        },
  {HTTP_400,  "Bad request"         },
  {HTTP_404,  "Not found"           },
  {HTTP_405,  "Method not allowed"  },
  {HTTP_500,  "Internal error"      },
  {HTTP_501,  "Not Implemented"     },
  {HTTP_503,  "Service Unavailable" },
  {0,0}
};

/*
 * CONSTANTS
 */

static char appName[] = "WiiBWerver/0.1";

// content types
static char defaultContentType[] = "text/html";
static char unknownContentType[] = "application/octet-stream";

// default template
static char error_template[] = "<html><head><title>Error #%d</title></head><body>The request has not been handled!<br/>%s</body></html>";

//---------------------------------------------------------------------
// Helper to read a http request.
//---------------------------------------------------------------------
char* read_full_request(int s, unsigned int *n) {
	char buf[512+1];
	int size = 512;
	int bcount = 0;		// bytes buffered
	int found = 0;		// end of request found
	int br;				// bytes read
	
	char* ptr = NULL;
	char* tmp = NULL;
	char* res = NULL;
	
	while(found == 0) {
		// read at most 512 chars
		if(-1 == (br = net_read(s, buf, 512))) {
			free(buf);
			*n = -1;
			return NULL;
		}
		buf[br] = '\0';

		for(ptr=&buf[0]; ptr<((char*)&buf[0])+br; ptr++) {
			// according to the spec, lines should end by '\r' '\n'
			if(*ptr == '\n') {
				// end of line found ...
				if(*(ptr+1) == '\r' && *(ptr+2) == '\n') {
					size = (int)(ptr+3 - buf);  // we should have (size==br)
					*n = bcount + size;
					found = 1;
					break;
				}
			}
		}
		// allocate a bigger buffer to continue
		tmp = (char*)malloc((bcount+size+1) * sizeof(char));
		memcpy(tmp, res, bcount);
		free(res);
		
		memcpy((char*)&tmp[bcount], buf, size);
		bcount += size;
		res = tmp;
		tmp = NULL;
	}
	
	// for pretty printing :-)
	res[*n] = '\0';	
	return res;	
}

//---------------------------------------------------------------------
// Helper to write data on the wire.
//---------------------------------------------------------------------
int write_data(int s, char *buf, int n) {
  int bcount;
  int br;
  bcount = 0;
  br = 0;
  
  ///printf("Writing %d data on the wire\n", n);
  
  while(bcount < n) {
    if ((br = net_write(s,buf,n-bcount)) > 0) {
      bcount += br;
      buf += br;
    }
    else if (br < 0)
      return br;
  }
  return (bcount);
}

//---------------------------------------------------------------------
// Sends the http status line. Returns the number of bytes writton on the wire.
//---------------------------------------------------------------------
int send_status_line(int socket, int code) {

	static char* no_message = "Unknown response";
	static char buffer[512];
	char *mesg = NULL;	
	int i, l;

	// try to find a message for the given http status code
	for(i=0; http_codes[i].code != 0; i++) {
		if(http_codes[i].code == code) {
			mesg = http_codes[i].message;
			break;
		}
	}
	if(mesg == NULL) mesg = no_message;

	l = sprintf(&buffer[0], "HTTP/1.0 %d %s\r\n", code, mesg);

	///printf("Status line: %s", buffer);

	return net_send(socket, buffer, l, 0);
}

//---------------------------------------------------------------------
// Sends common http headers. Returns the number of bytes writton on the wire.
//---------------------------------------------------------------------
int send_common_headers(int socket, int code, char* contentType, int length) {

	static char buffer[1024];	
	int l = 0;

	l += sprintf(&buffer[l], "Server: %s\r\n", appName); 
	l += sprintf(&buffer[l], "Content-Type: %s\r\n", contentType); 
	l += sprintf(&buffer[l], "Content-Length: %d\r\n",length); 
	l += sprintf(&buffer[l], "Connection: close\r\n"); 
	
	if(code == HTTP_405)
		l += sprintf(&buffer[l], "Allow: GET\r\n"); 
		
	l += sprintf(&buffer[l], "\r\n"); 

	///printf("Headers:\n"); printf(buffer);

	return net_send(socket, buffer, l, 0);
}

//---------------------------------------------------------------------
// Sends an error page to the client. Returns 0.
//---------------------------------------------------------------------
int send_error(int socket, int code, char* mesg) {

	char buffer[1024];	// should be enough ...
	int len;
	
	// TODO: try to load a template from sd card (in /wiibsrv/templates/)

	// compute human readable error message
	len = sprintf(&buffer[0], error_template, code, mesg);
	
	send_status_line(socket, code);
	send_common_headers(socket, code, defaultContentType, len);
	write_data(socket, (char*)&buffer[0], len);
	
	printf("Closing connection\n");
	(void)net_close(socket);
	return 0;
}

//---------------------------------------------------------------------
// Sends resource to the client. Returns 1 if everything goes well.
//---------------------------------------------------------------------
int send_resource(int socket, char* uri) {

	char path[2048];
	char *fstr = NULL;
	char *ptr = uri;
	int len, i, ret;
	
	size_t lSize, bSize;
	FILE* pFile;
	
	while(*ptr) {
		if(ptr-uri>1024) // URI is too long
			return send_error(socket, HTTP_500, "URI is too long (>1024)");
		ptr++;
	}
	
	// find file extension and thus content type
	while(*ptr!='.' && ptr>uri) ptr--;
	
	if(*ptr++ != '.') // no extension found, directory listing ?
		return send_error(socket, HTTP_501, "Directory listing is not yet implemented");
	
	// try to find a matching content-type
	for(i=0; extensions[i].ext != 0; i++) {
		len = strlen(extensions[i].ext);
		if(!strncmp(ptr, extensions[i].ext, len)) {
			fstr = extensions[i].filetype;
			break;
		}
	}
	if(fstr == NULL)
		fstr = unknownContentType;
						
	printf("Mime type is %s\n", fstr);
			
	sprintf(&path[0], "fat0:/wiibsrv/htdocs/%s", (*uri == '/' ? uri+1:uri));
	///printf("Opening %s\n", path);
	pFile = fopen (path, "r");
	if(pFile == NULL)
		return send_error(socket, HTTP_404, "Unable to open requested resource");
	
	// compute file size
	fseek(pFile , 0 , SEEK_END);
	lSize = ftell(pFile);
	rewind(pFile);
			
	///printf("File is %d bytes\n", lSize);
	
	// allocate a buffer to read the file (no memory waste :-))
	bSize = MIN(1024, lSize);
	if(bSize)
		ptr = malloc(bSize);
	
	send_status_line(socket, HTTP_200);
	send_common_headers(socket, HTTP_200, fstr, lSize);
	
	while(lSize>0) {
		///printf("Reading %d bytes from file\n", bSize);
		ret = fread(ptr, 1, bSize, pFile);
		write_data(socket, (char*)&ptr[0], ret);
		lSize -= ret;
	}
	if(lSize) {
		///printf("Reading %d bytes from file\n", lSize);
		ret = fread(ptr, 1, lSize, pFile);
		write_data(socket, (char*)&ptr[0], ret);
	}
	
	fclose(pFile);
	free(ptr);

	printf("Closing connection\n");
	(void)net_close(socket);
	return 1;
}

//---------------------------------------------------------------------
// the thread's main loop
//---------------------------------------------------------------------
void* process_request(void* param) {

	int socketfd = (int)param;
	char* request = NULL;
	unsigned int size;

	char *uri = NULL;			// pointer to the start of the uri
	char *ptr = NULL;
	
	// read the full incoming request
	///printf("Reading full request in a buffer\n");
	request = read_full_request(socketfd, &size);
	if(request == NULL) {
		printf("ARGGG: net_read() failed\n");
		(void)net_close(socketfd);
		return 0;
	}
	
	///printf("Received %d bytes\n", size);
	///printf(request); printf("\n");
					

	// check for valid http method
	if(strncmp(request,"GET ", 4) != 0)
		return (void*)send_error(socketfd, HTTP_405, "Only GET is supported");
	
	// find URI
	ptr = uri = (char*)&request[5];
	while(*ptr && *ptr != ' ') {
		if(*ptr == '.' && *(ptr+1) == '.')
			return (void*)send_error(socketfd, HTTP_400, "Parent directory (..) is not supported");
		ptr++;
	}
		
	// check status line
	if(*ptr == 0 || strncmp(ptr+1, "HTTP/1.", 7))
		return (void*)send_error(socketfd, HTTP_400, "Invalid status line");
	
	*ptr = '\0';

	send_resource(socketfd, uri);
	free(request);
	return (void*)1;
}

