#include <stdio.h>
#include <stdlib.h>
#include <gccore.h>
#include <fat.h>
#include <wiiuse/wpad.h>

#define THREADSTACK 8192

#include "..\include\server.h"

#define POWER_BUTTON    1
#define RESET_BUTTON 1<<1

static void *xfb = NULL;
static GXRModeObj *rmode = NULL;

static u8 frontButtons = 0;

//---------------------------------------------------------------------------------
// Called when the reset button is hit
//---------------------------------------------------------------------------------
static void reset_pushed() {
    frontButtons |= RESET_BUTTON;
}

//---------------------------------------------------------------------------------
// Called when the power button is hit
//---------------------------------------------------------------------------------
static void power_pushed() {
    frontButtons |= POWER_BUTTON;
}

//---------------------------------------------------------------------------------
// Reboot the wii
//---------------------------------------------------------------------------------
static void reboot() {
	int fd = IOS_Open("/dev/stm/immediate", 0);
	IOS_Ioctl(fd, 0x2001, NULL, 0, NULL, 0);
	IOS_Close(fd);
}

//---------------------------------------------------------------------------------
// Thread main routine
//---------------------------------------------------------------------------------
static void *threadone (void *arg) {
	printf("Starting web server from thread ...\n");
	run_server();
	return NULL;
}

//---------------------------------------------------------------------------------
// helper
//---------------------------------------------------------------------------------
static inline u32 wait_for(u32 btn)
{
	u32 btns;
	while(1)
	{
		WPAD_ScanPads();
		btns = WPAD_ButtonsDown(0);
		if(btns & btn)
			return btns;
		VIDEO_WaitVSync();
	}
}

static inline u32 wait_for_buttons()
{
	u32 btns;
	while(1)
	{
		WPAD_ScanPads();
		btns = WPAD_ButtonsDown(0);
		if(btns)
			return btns;
		VIDEO_WaitVSync();
	}
}

//---------------------------------------------------------------------------------
// main()
//---------------------------------------------------------------------------------
int main(int argc, char **argv) {

	lwp_t ThreadOneT;
	u32 btns;
	
	// set callbacks
	SYS_SetResetCallback(reset_pushed);
	SYS_SetPowerCallback(power_pushed);
	
	// very important to be able to access sd card
	fatInitDefault();
	
	// Initialise the video system
	VIDEO_Init();
	
	// This function initialises the attached controllers
	WPAD_Init();
	
	// Obtain the preferred video mode from the system
	// This will correspond to the settings in the Wii menu
	rmode = VIDEO_GetPreferredMode(NULL);

	// Allocate memory for the display in the uncached region
	xfb = MEM_K0_TO_K1(SYS_AllocateFramebuffer(rmode));
	
	// Initialise the console, required for printf
	console_init(xfb,20,20,rmode->fbWidth,rmode->xfbHeight,rmode->fbWidth*VI_DISPLAY_PIX_SZ);
	
	// Set up the video registers with the chosen mode
	VIDEO_Configure(rmode);
	
	// Tell the video hardware where our display memory is
	VIDEO_SetNextFramebuffer(xfb);
	
	// Make the display visible
	VIDEO_SetBlack(FALSE);

	// Flush the video register changes to the hardware
	VIDEO_Flush();

	// Wait for Video setup to complete
	VIDEO_WaitVSync();
	if(rmode->viTVMode&VI_NON_INTERLACE) VIDEO_WaitVSync();


	// The console understands VT terminal escape codes
	// This positions the cursor on row 2, column 0
	// we can use variables for this with format codes too
	// e.g. printf ("\x1b[%d;%dH", row, column );
	printf("\x1b[2;0H");
	
	printf("WiibServer by Jay.\n");	
	printf("Press A to start.\n");
	wait_for(WPAD_BUTTON_A);
	
	printf("Creating thread.\n");	
	LWP_CreateThread(&ThreadOneT, threadone, NULL, NULL, THREADSTACK, 80);
	printf("Thread created (%d), press B to quit.\n", ThreadOneT);	

	while(1) {

		// Call WPAD_ScanPads each loop, this reads the latest controller states
		WPAD_ScanPads();

		// WPAD_ButtonsDown tells us which buttons were pressed in this loop
		// this is a "one shot" state which will not fire again until the button has been released
		u32 pressed = WPAD_ButtonsDown(0);

		// We return to the launcher application via exit
		if ( pressed & WPAD_BUTTON_B ) break;

		// Wait for the next frame
		VIDEO_WaitVSync();
	}

	printf(" *** Press A to reboot now, B to exit to loader ***\n");
	while(1) {
		if(frontButtons & RESET_BUTTON)
			reboot();
			
		btns = wait_for_buttons();
		if(btns & WPAD_BUTTON_B) {
			return 0;
		}
		
		if(btns & WPAD_BUTTON_A)
			reboot();
 	}
	
	return 0;
}
